/**
 * Compile with: mex -v train_tree.cpp ShapeRFClassifier.cpp -I../../boost_1_54_0
 *
 * This code accompanies the paper:
 *
 * "Dense non-rigid shape correspondence using random forests"
 * E.Rodola, S.Rota Bulo, T.Windheuser, M.Vestner, D.Cremers.
 * Proc. CVPR 2014
 *
 * Please cite the paper above if you use the code.
 *
 * Emanuele Rodolà (c) 2014
 */
#include "mex.h"
#include <iostream>
#include "ShapeRFClassifier.h"
#include "mesh.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray*prhs[])
{
	if (nrhs != 4 || nlhs != 0)
      mexErrMsgTxt("Usage: train_tree(shapes, n_eigenvectors, n_energy_levels, out_filename).");

	const double* const models = mxGetPr(prhs[0]);
	const int n_models = int( mxGetN(prhs[0]) );
	
	const int n_eigenvectors = static_cast<int>(*mxGetPr(prhs[1]));
	const int n_energy_levels = static_cast<int>(*mxGetPr(prhs[2]));
	
	const std::string tree_fname = mxArrayToString(prhs[3]);
	
	if (n_models < 1)
		mexErrMsgTxt("At least one model shape should be used to train the forest.");
	
	if (n_eigenvectors < 2)
		mexErrMsgTxt("At least 2 eigenvectors are required.");
	
	if (n_energy_levels < 2)
		mexErrMsgTxt("At least 2 energy levels are required.");
		
	if (tree_fname.empty())
		mexErrMsgTxt("Output filename should be a valid path.");

	std::vector<std::string> model_prefix(n_models);

	for (int i=0; i<n_models; ++i)
	{
		mxArray* fld = mxGetFieldByNumber(prhs[0], 0, i);
		if (fld==0)
			mexErrMsgTxt("Error when parsing the models struct.");
		else
			model_prefix[i] = std::string(mxArrayToString(fld));
	}

	try
	{
		DataSet training_dataset(n_eigenvectors, n_energy_levels);
	  
		for (int m=0; m<n_models; ++m)
		{
			const std::string& prefix = model_prefix[m];
			int np;
			
			std::cout << "Adding training shape " << prefix << "... " << std::endl;
			
			std::vector<double> evals(n_eigenvectors);
			{
				const std::string fname = prefix + ".evals.dat";
				std::ifstream fin(fname.c_str(), std::ios::in);
				
				if (!fin.is_open())
				{
					const std::string msg = "Cannot open " + fname;
					mexErrMsgTxt(msg.c_str());
				}
					
				double tmp;
				fin >> tmp;
				np = static_cast<int>(tmp);
				
				for (int i=0; i<n_eigenvectors; ++i)
					fin >> evals[i];
				
				fin.close();
			}
			
			std::vector<double> evecs(np*n_eigenvectors);
			{
				const std::string fname = prefix + ".evecs.dat";
				std::ifstream fin(fname.c_str(), std::ios::in);
				
				if (!fin.is_open())
				{
					const std::string msg = "Cannot open " + fname;
					mexErrMsgTxt(msg.c_str());
				}
				
				for (size_t i=0; i<evecs.size(); ++i)
					fin >> evecs[i];
				
				fin.close();
			}
			
			// Load ground-truth correspondence
			
			std::vector<int> labels(np);
			std::vector<int> samples;
			samples.reserve(np);
			{
				const std::string fname = prefix + ".gt.txt";
				std::ifstream fin(fname.c_str(), std::ios::in);
				
				if (!fin.is_open())
				{
					const std::string msg = "Cannot open ground-truth labels " + fname;
					mexErrMsgTxt(msg.c_str());
				}
					
				for (size_t l=0; l<labels.size(); ++l)
				{
					fin >> labels[l];
					if (labels[l] > 0)
						samples.push_back(l);
				}
					
				fin.close();
			}

			training_dataset.addShape(labels, samples, evals, evecs);
		}

		ShapeRandomTree tree;
		tree.setDataSet(&training_dataset);
		tree.setNTrials(100);

		cout << "Training tree and saving to " << tree_fname << "..." << endl;
		tree.train();
        tree.save_binary(tree_fname);
        cout << "done." << endl;
	}
    catch (std::exception& e)
	{
		const std::string msg = "Exception caught: " + std::string(e.what());
		mexErrMsgTxt(msg.c_str());
	}
}
